using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using NUnit.Framework;
using System.Reflection;
using Xamarin.Utils;
using Xamarin.Tests;

namespace Xamarin.MMP.Tests {
	internal class MessageTool : Tool {
		public MessageTool ()
		{
		}

		public MessageTool (string text)
		{
			Output.Append (text);
		}

		protected override string ToolPath => throw new NotImplementedException ();

		protected override string MessagePrefix => "MM";
	}

	public class OutputText {
		public BuildResult BuildResult { get; private set; }
		public string RunOutput { get; private set; }

		public OutputText (BuildResult buildOutput, string runOutput)
		{
			BuildResult = buildOutput;
			RunOutput = runOutput;
		}

		internal MessageTool Messages {
			get {
				return BuildResult.Messages;
			}
		}
	}

	public class BuildResult {
		public readonly string BinLogPath;

		string build_output;
		public string BuildOutput {
			get {
				if (build_output is null)
					build_output = string.Join ("\n", BuildOutputLines);
				return build_output;
			}
		}

		string [] build_output_lines;
		public IList<string> BuildOutputLines {
			get {
				if (build_output_lines is null)
					build_output_lines = BinLog.PrintToLines (BinLogPath).ToArray ();
				return build_output_lines;
			}
		}

		public BuildResult (string binLogPath)
		{
			BinLogPath = binLogPath;
		}

		public bool HasMessage (int code)
		{
			return Messages.Messages.Any (v => v.Number == code);
		}

		public bool HasMessage (int code, string message)
		{
			return Messages.Messages.Any (v => v.Number == code && v.Message == message);
		}

		MessageTool messages;
		internal MessageTool Messages {
			get {
				if (messages is null) {
					messages = new MessageTool ();
					messages.ParseBinLog (BinLogPath);
				}
				return messages;
			}
		}
	}

	static class FrameworkBuilder {
		public static string CreateFatFramework (string tmpDir)
		{
			return Path.Combine (Configuration.RootPath, "tests", "test-libraries", "frameworks", ".libs", "MmpTestFramework.xcframework", "macos-arm64_x86_64", "MmpTestFramework.framework"); ;
		}

		public static string CreateThinFramework (string tmpDir, bool sixtyFourBits = true)
		{
			return Path.Combine (Configuration.RootPath, "tests", "test-libraries", "frameworks", ".libs", "osx-x64", "MmpTestFramework.framework");
		}
	}

	// Hide the hacks and provide a nice interface for writting tests that build / run XM projects
	static class TI {
		public class UnifiedTestConfig {
			public string TmpDir { get; set; }

			// Not necessarly required
			public bool FSharp { get; set; }
			public bool XM45 { get; set; }
			public bool Release { get; set; } = false;

			public string ProjectName { get; set; } = "";
			public string TestCode { get; set; } = "";
			public string TestDecl { get; set; } = "";
			public string CSProjConfig { get; set; } = "";
			public string References { get; set; } = "";
			public string ReferencesBeforePlatform { get; set; } = "";
			public string AssemblyName { get; set; } = "";
			public string ItemGroup { get; set; } = "";
			public string SystemMonoVersion { get; set; } = "";
			public string TargetFrameworkVersion { get; set; } = "";
			public Dictionary<string, string> PlistReplaceStrings { get; set; } = new Dictionary<string, string> ();
			public Tuple<string, string> CustomProjectReplacement { get; set; } = null;

			// Binding project specific
			public string APIDefinitionConfig { get; set; }
			public string StructsAndEnumsConfig { get; set; }
			public string LinkWithName { get; set; } = null; // Only generates if non-null

			// Unified Executable Specific
			public bool AssetIcons { get; set; }

			// Generated by TestUnifiedExecutable/TestSystemMonoExecutable and added to TestCode
			public Guid guid { get; set; }

			public string BundleName {
				get { return AssemblyName != "" ? AssemblyName : ProjectName.Split ('.') [0]; }
			}

			public string BundlePath {
				get { return Path.Combine (TmpDir, "bin", Release ? "Release" : "Debug", BundleName + ".app"); }
			}

			public string ExecutablePath {
				get { return Path.Combine (BundlePath, "Contents", "MacOS", BundleName); }
			}

			public UnifiedTestConfig (string tmpDir)
			{
				TmpDir = tmpDir;
			}
		}

		public static Version FindMonoVersion ()
		{
			string output = RunAndAssert ("/Library/Frameworks/Mono.framework/Commands/mono", new [] { "--version" }, "FindMonoVersion");

			Regex versionRegex = new Regex ("compiler version \\d+.\\d+.\\d+(.\\d+)?", RegexOptions.IgnoreCase);
			return new Version (versionRegex.Match (output).Value.Split (' ') [2]);
		}

		public static string RunAndAssert (string exe, params string [] args)
		{
			return RunAndAssert (exe, args, "Command: " + exe);
		}

		public static string RunAndAssert (string exe, IList<string> args, string stepName, bool shouldFail = false, Func<string> getAdditionalFailInfo = null, Dictionary<string, string> environment = null)
		{
			StringBuilder output = new StringBuilder ();
			environment ??= new Dictionary<string, string> ();
			environment ["MONO_PATH"] = null;
			environment ["DYLD_FALLBACK_LIBRARY_PATH"] = null;
			int compileResult = ExecutionHelper.Execute (exe, args, environmentVariables: environment, stdout: output, stderr: output);
			if (!shouldFail && compileResult != 0) {
				Console.WriteLine ($"Execution of the following command failed (exit code: {compileResult}):");
				Console.WriteLine ($"cd {Environment.CurrentDirectory}");
				foreach (var kvp in Environment.GetEnvironmentVariables ().Cast<System.Collections.DictionaryEntry> ().OrderBy (v => v.Key))
					Console.WriteLine ($"export {kvp.Key}={StringUtils.Quote (kvp.Value.ToString ())}");
				Console.WriteLine ($"{exe} {StringUtils.FormatArguments (args)}");
				Console.WriteLine (output);
			}
			Func<string> getInfo = () => getAdditionalFailInfo is not null ? getAdditionalFailInfo () : "";
			bool passed = shouldFail ? compileResult != 0 : compileResult == 0;
			if (!passed)
				Assert.Fail ($@"{stepName} {(shouldFail ? "passed" : "failed")} unexpectedly. Exit code: {compileResult}.");
			return output.ToString ();
		}

		// In most cases we generate projects in tmp and this is not needed. But nuget and test projects can make that hard
		public static void CleanUnifiedProject (string csprojTarget)
		{
			RunAndAssert (Configuration.XIBuildPath, new [] { "--", csprojTarget, "/t:clean" }, "Clean", environment: Configuration.GetBuildEnvironment (ApplePlatform.MacOSX));
		}

		public static BuildResult BuildProject (string csprojTarget, bool shouldFail = false, bool release = false, Dictionary<string, string> environment = null, IList<string> extraArgs = null)
		{
			Configuration.SetBuildVariables (ApplePlatform.MacOSX, ref environment);

			// This is to force build to use our mmp and not system mmp
			var buildArgs = new List<string> ();
			var binlog = Path.Combine (Path.GetDirectoryName (csprojTarget), $"log-{DateTime.Now:yyyyMMdd_HHmmss}.binlog");
			buildArgs.Add ($"/bl:{binlog}");
			Console.WriteLine ($"Binlog: {binlog}");

			// Restore any package references
			buildArgs.Add ("/r");

			if (release)
				buildArgs.Add ("/property:Configuration=Release");
			else
				buildArgs.Add ("/property:Configuration=Debug");

			if (extraArgs?.Count > 0)
				buildArgs.AddRange (extraArgs);

			buildArgs.Add (csprojTarget);

			Func<string> getBuildProjectErrorInfo = () => {
				string csprojText = "\n\n\n\tCSProj: \n" + File.ReadAllText (csprojTarget);
				string csprojLocation = Path.GetDirectoryName (csprojTarget);
				string fileList = "\n\n\tFiles: " + String.Join (" ", Directory.GetFiles (csprojLocation).Select (x => x.Replace (csprojLocation + "/", "")));
				return csprojText + fileList;
			};

			buildArgs.Insert (0, "--");
			RunAndAssert (Configuration.XIBuildPath, buildArgs, "Compile", shouldFail, getBuildProjectErrorInfo, environment);

			return new BuildResult (binlog);
		}

		static string ProjectTextReplacement (UnifiedTestConfig config, string text)
		{
			text = text.Replace ("REPLACE_CODE_REPLACE", config.CSProjConfig)
					   .Replace ("%REFERENCES%", config.References)
					   .Replace ("%REFERENCES_BEFORE_PLATFORM%", config.ReferencesBeforePlatform)
					   .Replace ("%NAME%", config.AssemblyName ?? Path.GetFileNameWithoutExtension (config.ProjectName))
					   .Replace ("%ITEMGROUP%", config.ItemGroup)
					   .Replace ("%TARGET_FRAMEWORK_VERSION%", config.TargetFrameworkVersion);
			if (config.CustomProjectReplacement is not null)
				text = text.Replace (config.CustomProjectReplacement.Item1, config.CustomProjectReplacement.Item2);
			return text;
		}

		public static string RunEXEAndVerifyGUID (string tmpDir, Guid guid, string path)
		{
			// Assert that the program actually runs and returns our guid
			Assert.IsTrue (File.Exists (path), string.Format ("{0} did not generate an exe?", path));
			string output = RunAndAssert (path, Array.Empty<string> (), "Run");

			string guidPath = Path.Combine (tmpDir, guid.ToString ());
			Assert.IsTrue (File.Exists (guidPath), "Generated program did not create expected guid file: " + output);

			// Let's delete the guid file so re-runs inside same tests are accurate
			File.Delete (guidPath);
			return output;
		}

		public static string GenerateEXEProject (UnifiedTestConfig config)
		{
			WriteMainFile (config.TestDecl, config.TestCode, config.FSharp, Path.Combine (config.TmpDir, config.FSharp ? "Main.fs" : "Main.cs"));

			string sourceDir = FindSourceDirectory ();

			if (config.AssetIcons) {
				RunAndAssert ("/bin/cp", new [] { "-R", Path.Combine (sourceDir, "Icons/Assets.xcassets"), config.TmpDir }, "Copy Asset Icons");
				config.ItemGroup += @"<ItemGroup>
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\Contents.json"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-128.png"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-128%402x.png"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-16.png"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-16%402x.png"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-256%402x.png"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-32.png"" />
    <ImageAsset Include=""Assets.xcassets\AppIcon.appiconset\AppIcon-32%402x.png"" />
    <ImageAsset Include=""Assets.xcassets\Contents.json"" />
  </ItemGroup>";
				// HACK - Should process using CopyFileWithSubstitutions
				config.PlistReplaceStrings.Add ("</dict>", @"<key>XSAppIconAssets</key><string>Assets.xcassets/AppIcon.appiconset</string></dict>");
			}

			CopyFileWithSubstitutions (Path.Combine (sourceDir, "Info-Unified.plist"), Path.Combine (config.TmpDir, "Info.plist"), text => {
				foreach (var key in config.PlistReplaceStrings.Keys)
					text = text.Replace (key, config.PlistReplaceStrings [key]);

				return text;
			});

			return CopyFileWithSubstitutions (Path.Combine (sourceDir, config.ProjectName), Path.Combine (config.TmpDir, config.ProjectName), text => {
				return ProjectTextReplacement (config, text);
			});
		}

		public static string GenerateBindingLibraryProject (UnifiedTestConfig config)
		{
			string sourceDir = FindSourceDirectory ();
			CopyFileWithSubstitutions (Path.Combine (sourceDir, "ApiDefinition.cs"), Path.Combine (config.TmpDir, "ApiDefinition.cs"), text => text.Replace ("REPLACE_CODE_REPLACE", config.APIDefinitionConfig));
			CopyFileWithSubstitutions (Path.Combine (sourceDir, "StructsAndEnums.cs"), Path.Combine (config.TmpDir, "StructsAndEnums.cs"), text => text.Replace ("REPLACE_CODE_REPLACE", config.StructsAndEnumsConfig));

			string linkWithName = null;
			if (config.LinkWithName is not null) {
				string fileName = Path.GetFileNameWithoutExtension (config.LinkWithName);
				linkWithName = $"{fileName}.linkwith.cs";
				File.WriteAllText (Path.Combine (config.TmpDir, linkWithName), $@"using ObjCRuntime;

[assembly: LinkWith (""{config.LinkWithName}"", SmartLink = true, ForceLoad = true)]");

			}

			return CopyFileWithSubstitutions (Path.Combine (sourceDir, config.ProjectName), Path.Combine (config.TmpDir, config.ProjectName), text => {
				if (linkWithName is not null)
					text = text.Replace ("%ITEMGROUP%", $@"<ItemGroup><Compile Include=""{linkWithName}"" /></ItemGroup>%ITEMGROUP%");
				return ProjectTextReplacement (config, text);
			});
		}

		public static string GenerateUnifiedLibraryProject (UnifiedTestConfig config)
		{
			string sourceDir = FindSourceDirectory ();
			string sourceFileName = config.FSharp ? "Component1.fs" : "MyClass.cs";
			string projectSuffix = config.FSharp ? ".fsproj" : ".csproj";

			CopyFileWithSubstitutions (Path.Combine (sourceDir, sourceFileName), Path.Combine (config.TmpDir, sourceFileName), text => {
				return text.Replace ("REPLACE_CODE_REPLACE", config.TestCode);
			});

			return CopyFileWithSubstitutions (Path.Combine (sourceDir, config.ProjectName + projectSuffix), Path.Combine (config.TmpDir, config.ProjectName + projectSuffix), text => {
				return ProjectTextReplacement (config, text);
			});
		}

		public static string GenerateNetStandardProject (UnifiedTestConfig config)
		{
			const string SourceFile = "Class1.cs";
			const string ProjectFile = "NetStandardLib.csproj";
			const string NetStandardSubDir = "NetStandard";

			string sourceDir = FindSourceDirectory ();

			Directory.CreateDirectory (Path.Combine (config.TmpDir, NetStandardSubDir));
			File.Copy (Path.Combine (sourceDir, NetStandardSubDir, SourceFile), Path.Combine (config.TmpDir, NetStandardSubDir, SourceFile), true);

			string projectPath = Path.Combine (config.TmpDir, NetStandardSubDir, ProjectFile);
			File.Copy (Path.Combine (sourceDir, NetStandardSubDir, ProjectFile), projectPath, true);

			return projectPath;
		}

		public static string GetUnifiedExecutableProjectName (UnifiedTestConfig config)
		{
			string projectName;
			if (config.FSharp)
				projectName = config.XM45 ? "FSharpXM45Example" : "FSharpUnifiedExample";
			else
				projectName = config.XM45 ? "XM45Example" : "UnifiedExample";
			string projectExtension = config.FSharp ? ".fsproj" : ".csproj";
			return projectName + projectExtension;
		}

		public static string GenerateUnifiedExecutableProject (UnifiedTestConfig config)
		{
			config.ProjectName = GetUnifiedExecutableProjectName (config);
			return GenerateEXEProject (config);
		}

		public static BuildResult GenerateAndBuildUnifiedExecutable (UnifiedTestConfig config, bool shouldFail = false, Dictionary<string, string> environment = null)
		{
			string csprojTarget = GenerateUnifiedExecutableProject (config);
			return BuildProject (csprojTarget, shouldFail: shouldFail, release: config.Release, environment: environment);
		}

		public static string RunGeneratedUnifiedExecutable (UnifiedTestConfig config)
		{
			return RunEXEAndVerifyGUID (config.TmpDir, config.guid, config.ExecutablePath);
		}

		public static OutputText TestUnifiedExecutable (UnifiedTestConfig config, bool shouldFail = false, Dictionary<string, string> environment = null)
		{
			AddGUIDTestCode (config);

			var buildOutput = GenerateAndBuildUnifiedExecutable (config, shouldFail, environment);
			if (shouldFail)
				return new OutputText (buildOutput, "");

			string runOutput = RunGeneratedUnifiedExecutable (config);

			return new OutputText (buildOutput, runOutput);
		}

		public static void AddGUIDTestCode (UnifiedTestConfig config)
		{
			// If we've already generated guid bits for this config, don't tack on a second copy
			if (config.guid == Guid.Empty) {
				config.guid = Guid.NewGuid ();
				config.TestCode += GenerateOutputCommand (config.TmpDir, config.guid);
			}
		}

		public static OutputText TestSystemMonoExecutable (UnifiedTestConfig config, bool shouldFail = false)
		{
			config.guid = Guid.NewGuid ();
			var projectName = "SystemMonoExample";
			config.TestCode += GenerateOutputCommand (config.TmpDir, config.guid);
			config.ProjectName = $"{projectName}.csproj";
			string csprojTarget = GenerateSystemMonoEXEProject (config);

			var buildOutput = BuildProject (csprojTarget, shouldFail: shouldFail, release: config.Release);
			if (shouldFail)
				return new OutputText (buildOutput, "");

			string exePath = Path.Combine (config.TmpDir, "bin", config.Release ? "Release" : "Debug", projectName + ".app", "Contents", "MacOS", projectName);
			string runOutput = RunEXEAndVerifyGUID (config.TmpDir, config.guid, exePath);
			return new OutputText (buildOutput, runOutput);
		}

		static string GetTargetFrameworkValue (UnifiedTestConfig config)
		{
			string version = config.SystemMonoVersion == "" ? "4.5" : config.SystemMonoVersion;
			return string.Format ("<TargetFrameworkVersion>v{0}</TargetFrameworkVersion>", version);
		}

		public static string GenerateSystemMonoEXEProject (UnifiedTestConfig config)
		{
			WriteMainFile (config.TestDecl, config.TestCode, false, Path.Combine (config.TmpDir, "Main.cs"));

			string sourceDir = FindSourceDirectory ();
			File.Copy (Path.Combine (sourceDir, "Info-Unified.plist"), Path.Combine (config.TmpDir, "Info.plist"), true);

			return CopyFileWithSubstitutions (Path.Combine (sourceDir, config.ProjectName), Path.Combine (config.TmpDir, config.ProjectName), text => {
				return ProjectTextReplacement (config, text.Replace ("%TARGETFRAMEWORKVERSION%", GetTargetFrameworkValue (config)));
			});
		}

		public static string TestDirectory => Path.Combine (Configuration.RootPath, "tests");

		public static string FindSourceDirectory ()
		{
			return Path.Combine (TestDirectory, "common", "mac");
		}

		public static void CopyDirectory (string src, string target)
		{
			Assert.AreEqual (0, ExecutionHelper.Execute ("/bin/cp", new [] { "-r", src, target }));
		}

		public static string CopyFileWithSubstitutions (string src, string target, Func<string, string> replacementAction)
		{
			string text = replacementAction (System.IO.File.ReadAllText (src));
			System.IO.File.WriteAllText (target, text);
			return target;
		}

		static void WriteMainFile (string decl, string content, bool fsharp, string location)
		{
			const string FSharpMainTemplate = @"
namespace FSharpUnifiedExample
open System
open AppKit

module main =
    %DECL%
 
    [<EntryPoint>]
    let main args =
        NSApplication.Init ()
        REPLACE_CODE_REPLACE
        0";

			const string MainTemplate = @"
using Foundation;
using AppKit;

namespace TestCase
{
	class MainClass
	{
		%DECL%

		static void Main (string[] args)
		{
			NSApplication.Init ();
			REPLACE_CODE_REPLACE
		}
	}
}";
			string currentTemplate = fsharp ? FSharpMainTemplate : MainTemplate;
			string testCase = currentTemplate.Replace ("REPLACE_CODE_REPLACE", content).Replace ("%DECL%", decl);
			using (StreamWriter s = new StreamWriter (location))
				s.Write (testCase);
		}

		public static string FindRootDirectory ()
		{
			return Configuration.TargetDirectoryXM;
		}

		static string GenerateOutputCommand (string tmpDir, Guid guid)
		{
			return string.Format ("System.IO.File.Create(\"{0}\").Dispose();", Path.Combine (tmpDir, guid.ToString ()));
		}

		public static bool TryNugetRestore (string project, out StringBuilder output)
		{
			output = new StringBuilder ();
			var rv = ExecutionHelper.Execute (Configuration.XIBuildPath, new [] { $"--", "/t:Restore", project }, stdout: output, stderr: output, environmentVariables: Configuration.GetBuildEnvironment (ApplePlatform.MacOSX));
			return rv == 0;
		}

		public static void NugetRestore (string project)
		{
			if (!TryNugetRestore (project, out var output)) {
				Console.WriteLine ("nuget restore failed:");
				Console.WriteLine (output);
				Assert.Fail ($"'nuget restore' failed for {project}");
			}
		}

		public static bool InJenkins {
			get {
				var buildRev = Environment.GetEnvironmentVariable ("BUILD_REVISION");
				return !string.IsNullOrEmpty (buildRev) && buildRev == "jenkins";
			}
		}
	}

	static class PlatformHelpers {
		// Yes, this is a copy of the one in PlatformAvailability.cs. However, right now
		// we don't depend on Xamarin.Mac.dll, so moving to it was too painful. If we start
		// using XM, we can revisit.
		const int sys1 = 1937339185;
		const int sys2 = 1937339186;

		// Deprecated in OSX 10.8 - but no good alternative is (yet) available
		[System.Runtime.InteropServices.DllImport ("/System/Library/Frameworks/Carbon.framework/Versions/Current/Carbon")]
		static extern int Gestalt (int selector, out int result);

		static int osx_major, osx_minor;

		public static bool CheckSystemVersion (int major, int minor)
		{
			if (osx_major == 0) {
				Gestalt (sys1, out osx_major);
				Gestalt (sys2, out osx_minor);
			}
			return osx_major > major || (osx_major == major && osx_minor >= minor);
		}
	}
}
